"""Package to inspect, manage and modify block devices."""

from __future__ import annotations
import types
import pathlib
import math

from blkmgr import wrapper


def get_disks() -> list[Disk] :
    """Get list of all disks available to this system.

    Returns:
        List of disks on this system.
    """
    disks = []

    listing = wrapper.lsblk(['path', 'type'])
    for block_data in listing :
        if block_data['type'] == Disk.REF_TYPE :
            disks.append(Disk(block_data['path']))
            pass
        pass

    return disks


class Block :
    """Block device super class."""

    REF_TYPE = "block"
    """Reference string for block type."""

    @classmethod
    def init_kname(cls, kname) :
        """initialize block device by given kernel name."""
        query = wrapper.lsblk(output=["path", "kname"], dependants=True)
        for device in query :
            if device["kname"] == kname :
                return cls(device['path'])
        return None

    def __init__(self, path:str) :
        """
        Parameters:
            path :
                String of path to device node.
        """

        self.device_path:str = path
        """Device path of this block device."""

        return

    def __str__(self) :
        return self.__repr__()
    def __repr__(self) :
        return f"{type(self).__name__}({self.name})"

    def lsblk_get(self,output:str) -> str|None :
        """Get lsblk property

        Helper function to request single lsblk property.

        Parameters:
            output :
                String of key to output.

        Returns:
            Requested lsblk property value.
        """
        return wrapper.lsblk(output=[output], device=self.device_path)[0][output]

    def partprobe(self) :
        """Run partptobe on this device."""
        wrapper.partprobe(self.device_path)
        return

    @property
    def name(self) -> str :
        """Name of this block device."""
        return self.lsblk_get("name")

    @property
    def kernel_name(self) -> str :
        """Kernel name of this device."""
        return self.lsblk_get("kname")

    @property
    def parent_kernel_name(self) -> str :
        """Kernel name of parent device."""
        return self.lsblk_get("pkname")

    @property
    def type_name(self) -> str :
        """Name of block device type."""
        return self.lsblk_get("type")

    @property
    def device_numbers(self) -> tuple[str, str] :
        """Tuple of (major, minor) device numbers."""
        return tuple([int(n) for n in self.lsblk_get("maj:min").split(":")])

    @property
    def file_system_type_name(self) -> str :
        """Name of file system type."""
        return self.lsblk_get("fstype")

    @property
    def partition_table_type_name(self) -> str :
        """Name of partition table type."""
        return self.lsblk_get("pttype")

    @property
    def size(self) -> int :
        """Block device size in bytes."""
        return int(self.lsblk_get("size"))

    @property
    def phy_sec_size(self) -> int :
        """Physical sector size in bytes."""
        return int(self.lsblk_get("phy-sec"))

    @property
    def log_sec_size(self) -> int :
        """Logical sector size in bytes."""
        return int(self.lsblk_get("log-sec"))

    def list_dependant_paths(self) -> list[str] :
        """Returns list of all dependant device paths."""

        # Query device dependants.
        dependants_query = wrapper.lsblk(output=["path", "pkname"], device=self.device_path, dependants=True)

        # List for matching dependants.
        dependants = []

        # Verify dependant parent kernel name.
        for dependant in dependants_query :
            if dependant["pkname"] == self.kernel_name :
                dependants.append(dependant["path"])
            pass

        return dependants

    def byte_to_sec(self, b:int) -> int :
        """Convert bytes to sectors. Note: always rounds up if result has decimals."""
        return math.ceil(b / self.log_sec_size)

    def kib_to_sec(self, kib:int|float) -> int :
        """Convert kibibyte size to sector size."""
        return self.byte_to_sec((1024 ** 1) * kib)

    def mib_to_sec(self, mib) :
        """Convert mebibibyte size to sector size."""
        return self.byte_to_sec((1024 ** 2) * mib)

    def gib_to_sec(self, gib) :
        """Convert gigibyte size to sector size."""
        return self.byte_to_sec((1024 ** 3) * gib)

    def tib_to_sec(self, tib) :
        """Convert tebibyte size to sector size."""
        return self.byte_to_sec((1024 ** 4) * tib)

    def sec_to_byte(self, sec:int) -> int :
        """Convert sectors to bytes."""
        return int(sec * self.log_sec_size)

    def sec_to_kib(self, sec:int) -> float :
        """Converts sectors to kibibytes."""
        return float(self.sec_to_byte(sec) / (1024 ** 1))

    def sec_to_mib(self, sec:int) -> float :
        """Converts sectors to mebibytes."""
        return float(self.sec_to_byte(sec) / (1024 ** 2))

    def sec_to_gib(self, sec:int) -> float :
        """Converts sectors to gigibytes."""
        return float(self.sec_to_byte(sec) / (1024 ** 3))

    def sec_to_tib(self, sec:int) -> float :
        """Converts sectors to tebibytes."""
        return float(self.sec_to_byte(sec) / (1024 ** 4))

    pass


class Disk (Block) :
    """Disk block device"""

    REF_TYPE = "disk"

    def __init__(self, *args, **kwargs) :
        super().__init__(*args, **kwargs)
        return

    @property
    def partitions(self) -> list[Partition] :
        """List of initialized partitions on this disk"""
        partitions = []
        for dependant_path in self.list_dependant_paths() :
            partitions.append(Partition(dependant_path))
            pass
        return partitions

    @property
    def first_in_largest_sec(self) -> int :
        """Returns first sector in largest unallocated space on this disk."""
        return wrapper.sgdisk_first_in_largest(self.device_path)

    @property
    def first_in_largest_aligned_sec(self) -> int :
        """Returns first aligned sector in largest unallocated space on this disk."""
        return wrapper.sgdisk_first_aligned_in_largest(self.device_path)

    @property
    def last_in_largest_sec(self) -> int :
        """Returns last sector in largest unallocated space on this disk."""
        return wrapper.sgdisk_end_of_largest(self.device_path)

    def wipe(self) :
        """wipe disk partition table."""
        wrapper.sgdisk_zap(self.device_path)
        return

    def chptt_gpt(self) :
        """Change partition table type to GPT."""
        wrapper.parted_mklabel(self.device_path, 'gpt')
        return

    def create_partition_sec(self, start_sector, end_sector) :
        """Create new partition from to given sectors."""
        start_paths = self.list_dependant_paths()
        wrapper.parted_mkpart(self.device_path, start_sector, end_sector, unit='s')
        finish_paths = self.list_dependant_paths()
        diff_paths = list(set(start_paths) ^ set(finish_paths))
        if len(diff_paths) == 1 :
            return Partition(diff_paths[0])
        return None

    def create_partition_first_sec(self, sector_size, align=True) :
        """Create partition of given sector size starting from first possible sector."""
        start_sector = self.first_in_largest_sec
        if align :
            start_sector = self.first_in_largest_aligned_sec
            pass
        end_sector = start_sector + sector_size
        return self.create_partition_sec(start_sector, end_sector)

    def create_partition_last_sec(self, sector_size) :
        """Create partition of given sector size ending at last possible sector."""
        end_sector = self.last_in_largest_sec
        start_sector = end_sector - sector_size
        return self.create_partition_sec(start_sector, end_sector)

    def create_partition_start_sec_fill(self, start_sector) :
        """Create partition from given start and fill all available space."""
        end_sector = self.last_in_largest_sec
        return self.create_partition_sec(start_sector, end_sector)

    def create_partition_end_sec_fill(self, end_sector, align=True) :
        """Create partition filling all available space until end sector."""
        start_sector = self.first_in_largest_sec
        if align :
            start_sector = self.first_in_largest_aligned_sec
            pass
        return self.create_partition_sec(start_sector, end_sector)

    def create_partition_fill(self, align=True) :
        """Create partition filling largest available space."""
        start_sector = self.first_in_largest_sec
        if align :
            start_sector = self.first_in_largest_aligned_sec
            pass
        end_sector = self.last_in_largest_sec
        return self.create_partition_sec(start_sector, end_sector)

    pass



# [i] Represents a partition.
class Partition (Block) :
    """Partition block device."""

    REF_TYPE = "part"

    @classmethod
    def init_part_label(cls, part_label) :
        """Initializes first partition matching given partition label."""
        query = wrapper.lsblk(output=["path", "partlabel"], dependants=True)
        for device in query :
            if device["partlabel"] == part_label :
                return cls(device['path'])
        return None

    def __init__(self, *args, **kwargs) :
        super().__init__(*args, **kwargs)
        return

    @property
    def disk(self) -> Disk :
        """Related parent disk instance."""
        return Disk.init_kname(self.parent_kernel_name)

    @property
    def part_number(self) :
        """Number of this partition."""
        return self.device_numbers[1]

    @property
    def part_label(self) :
        """Partition label."""
        return self.lsblk_get('partlabel')
    @part_label.setter
    def part_label(self, new_label) :
        wrapper.parted_name(self.disk.device_path, self.part_number, new_label)
        self.partprobe()
        return

    @property
    def part_uuid(self) :
        """Partition UUID. Random generated if set to `None`."""
        return self.lsblk_get('partuuid')
    @part_uuid.setter
    def part_uuid(self, new_uuid) :
        wrapper.sgdisk_partition_guid(self.disk.device_path, self.part_number, new_uuid)
        self.partprobe()
        return

    @property
    def typecode(self) :
        """Partition type code."""
        return self.lsblk_get('parttype')
    @typecode.setter
    def typecode(self, typecode) :
        if typecode is None :
            typecode = '00000000-0000-0000-0000-000000000000'
            pass
        wrapper.sgdisk_typecode(self.disk.device_path, self.part_number, typecode)
        self.partprobe()
        return

    @property
    def mountpath(self) :
        """Path of partition mountpoint. Set to `None` to unmount."""
        mountpoint = self.lsblk_get('mountpoint')
        if mountpoint is None :
            return None
        return pathlib.Path(mountpoint).resolve()
    @mountpath.setter
    def mountpath(self, mountpath) :
        if mountpath is None :
            wrapper.umount(self.device_path)
            self.partprobe()
            return
        if self.mountpath is not None :
            wrapper.umount(self.device_path)
            self.partprobe()
            pass
        wrapper.mount(self.device_path, mountpath)
        self.partprobe()
        return

    def chfst_ext4(self) :
        """Change file system type to ext4."""
        wrapper.mkfs_ext4(self.device_path)
        self.partprobe()
        return

    def chfst_vfat(self) :
        """Change file system type to vfat."""
        wrapper.mkfs_vfat(self.device_path)
        self.partprobe()
        return

    def chfst_btrfs(self) :
        """Change file system type to btrfs."""
        wrapper.mkfs_btrfs(self.device_path)
        self.partprobe()
        return

    def chfst_swap(self) :
        """Change file system type to linux swap."""
        wrapper.mkswap(self.device_path)
        self.partprobe()
        return

    pass





"""Wrapper functions for blkmgr."""


import subprocess
import json
import pathlib


def _sp_log(p) :
    """Log subprocess return for development."""
    print('*'*80)
    print('ARGS:', p.args)
    print('STDOUT:', p.stdout)
    print('STDERR:', p.stderr)
    print('*'*80)
    return

def lsblk(output:list[str]=[], device:str=None, dependants:bool=False) -> dict[str] :
    """lsblk Wrapper

    Performs a system call to lsblk to retrieve requested data.

    Parameters:
        output:
            List of strings defining requested keys. If empty, all possible outputs will be listed.
        device:
            String name of a specific device to request. None lists all devices.
        dependants:
            If dependant devices should be listed.

    Returns:
        List of block device information dictionaries.
    """

    # Base lsblk parameters.
    p_args = ["lsblk", "--json", "--bytes"]

    # List specific outputs or all.
    if len(output) > 0 :
        p_args.append(f"--output={','.join(output)}")
        pass
    else :
        p_args.append("--output-all")
        pass

    # Do not list dependants.
    if not dependants :
        p_args.append("--nodeps")
        pass

    # Specific requested device.
    if device is not None :
        p_args.append(device)
        pass

    # Perform lsblk system call.
    p = subprocess.run(p_args, capture_output=True)
    #_sp_log(p)

    # Process and return block device information.
    lsblk_out = json.loads(p.stdout)
    return lsblk_out["blockdevices"]

def partprobe(device=None) :
    """Inform kernel of partition table changes.

    Parameters:
        device:
            Device path to partprobe only a specific device.
    """
    p_cmd = ["partprobe"] #, "--summary"]
    if device is not None :
        p_cmd.append(device)
        pass
    p = subprocess.run(p_cmd, capture_output=True)
    #_sp_log(p)
    return

def sgdisk_zap(device:str) :
    """Destroy partition table data structure.

    Parameters:
        device:
            Device path of disk to zap."""
    p_args = ['sgdisk', '--zap-all', device]
    p = subprocess.run(p_args, capture_output=False)
    #_sp_log(p)
    return

def sgdisk_first_in_largest(device:str) -> int :
    """First sector in largest unallocated space on disk.

    Parameters:
        device:
            Device to get first in largest sector of.
    Returns:
        First available sector on disk.
    """
    p_cmd = ['sgdisk', '--first-in-largest', device]
    p = subprocess.run(p_cmd, capture_output=True)
    #_sp_log(p)
    return int(p.stdout)

def sgdisk_first_aligned_in_largest(device:str) -> int :
    """First aligned sector in largest unallocated space on disk.

    Parameters:
        device:
            Device to get first in largest sector of.
    Returns:
        First available sector on disk.
    """
    p_cmd = ['sgdisk', '--first-aligned-in-largest', device]
    p = subprocess.run(p_cmd, capture_output=True)
    #_sp_log(p)
    return int(p.stdout)

def sgdisk_end_of_largest(device:str) -> int :
    """Last sector of largest unallocated space on disk.

    Parameters:
        device:
            Device to get end of largest sector of.
    Returns:
        First available sector on disk.
    """
    p_cmd = ['sgdisk', '--end-of-largest', device]
    p = subprocess.run(p_cmd, capture_output=True)
    #_sp_log(p)
    return int(p.stdout)

def sgdisk_partition_guid(disk:str, partition:int, guid:str=None) :
    """Set GUID on given partition.

    Parameters:
        disk:
            Disk containing partition to set GUID on.
        partition:
            Number of partition to set GUID on.
        guid:
            GUID to set or `None` if it should be generated.
    """
    if guid is None :
        guid = 'R'
        pass
    p_cmd = ['sgdisk', f'--partition-guid={partition}:{guid}', disk]
    p = subprocess.run(p_cmd, capture_output=True)
    #_sp_log(p)
    return

def sgdisk_disk_guid(disk:str, guid:str=None) :
    """Set GUID on given disk.

    Parameters:
        disk:
            Disk containing partition to set GUID on.
        guid:
            GUID to set or `None` if it should be generated.
    """
    if guid is None :
        guid = 'R'
        pass
    p_cmd = ['sgdisk', f'--disk-guid={guid}', disk]
    p = subprocess.run(p_cmd)
    #_sp_log(p)
    return

def sgdisk_typecode(disk:str, partition:int, typecode:str) :
    """Set typecode (hex or GUID) of a partition.

    Parameters:
        disk:
            Disk containing partition to set typecode on.
        partition:
            Number of partition to set typecode on.
        typecode:
            typecode as str(hex) or guid.
    """
    p_cmd = ['sgdisk', f'--typecode={partition}:{typecode}', disk]
    p = subprocess.run(p_cmd, capture_output=True)
    #_sp_log(p)
    return

def parted_mklabel(disk:str, label_type:str) :
    """Create new disk table of given type.

    Parameters:
        device:
            Device path to create new partition table type on.
        label_type:
            Partition table type name to create.
            Choices: `bsd`, `dvh`, `gpt`, `loop`, `mac`, `msdos`, `pc98`, `sun`
    """
    p_args = ['parted', '--script', disk, 'mklabel', label_type]
    p = subprocess.run(p_args, capture_output=True)
    #_sp_log(p)
    return

def parted_mkpart(device:str, start:int, end:int, unit:str='B', part_type='primary') :
    """Create new partition.

    Parameters:
        device:
            Device path of disk to create partition on.
        start:
            Start of partition.
        end:
            End of partition.
        unit:
            Units used for start and end.
            Choices: `s` (sectors), `B` (bytes), `kB`, `MB`, `GB`, `TB`, `%` (of disk device)
        part_type:
            Type of new partition.
    """
    p_cmd = ['parted', '--script', device,
        'unit', unit,
        'mkpart', part_type, str(start), str(end)]
    p = subprocess.run(p_cmd, capture_output=True)
    #_sp_log(p)
    return

def parted_name(disk, partition, name) :
    """Set a partition name.

    Parameters:
        disk:
            Device path of disk to rename partition of.
        partition:
            Number of partition to rename.
        name:
            New name of partition.
    """
    p_cmd = ['parted', '--script', disk,
        'name', str(partition), name]
    p = subprocess.run(p_cmd, capture_output=True)
    #_sp_log(p)
    return

def mkfs_ext4(device) :
    """Create EXT4 file system on given device."""
    p_args = ["mkfs.ext4", device]
    p = subprocess.run(p_args, capture_output=True)
    #_sp_log(p)
    return

def mkfs_vfat(device, fat_size=32) :
    """Create FAT32 file system on given device."""
    p_args = ["mkfs.vfat", "-F", str(fat_size), device]
    p = subprocess.run(p_args, capture_output=True)
    #_sp_log(p)
    return

def mkfs_btrfs(device) :
    """Create BTRFS file system on given device."""
    p_args = ["mkfs.btrfs", device]
    p = subprocess.run(p_args, capture_output=True)
    #_sp_log(p)
    return

def mkswap(device) :
    """Create Linux SWAP file system on given device."""
    p_args = ["mkswap", device]
    p = subprocess.run(p_args, capture_output=True)
    #_sp_log(p)
    return

def mount(device:str, path) :
    """Mount device to given path.

    Parameters:
        device:
            Device path to mount.
        path :
            Directory to mount to. Will be created if not existing.
    """
    path = pathlib.Path(path).resolve()
    path.mkdir(parents=True, exist_ok=True)
    p_args = ["mount", device, path]
    p = subprocess.run(p_args, capture_output=True)
    #_sp_log(p)
    return

def umount(device:str) :
    """Unmount device or path.

    Parameters:
        device:
            Device path or dir path to unmount.
    """
    p_args = ["umount", device]
    p = subprocess.run(p_args, capture_output=True)
    #_sp_log(p)
    return

